//
// Project : webgen
// File    : main.cc
// Purpose : main-function
// License : GPL
//
// arch-tag: 259c11ec-c8b4-4a9d-b800-2c474635d332
//

#include <stdio.h>
#include <sys/types.h>
#include <time.h>

#include <fstream>
#include <map>

#include "TSLL.hh"
#include "Menu.hh"
#include "parser.hh"

using std::cout;
using std::endl;
using std::ostream;
using std::ofstream;
using std::ifstream;
using std::string;
using std::map;

// forward decl.
static void   parse_cmd     ( int argc, char ** argv );
static void   usage         ( int status, const string & param = "" );
static void   print_node    ( const string & id );
static void   print_node    ( Menu * menu );
static void   print_menu    ( ostream & os, Menu * node, Menu * outnode, string ofs );
static void   print_tabs    ( ostream & os, Menu * node, Menu * outnode, string ofs );
static void   print_content ( ostream & os, Menu * node );
static void   print_header  ( ostream & os, Menu * node );
static void   print_footer  ( ostream & os, Menu * node );
static string rel_link      ( const string & file1, const string & file2 );
static string basedir       ( string & file );
static void   adjust_sub    ( Menu * father, Menu * sub );
static void   adjust_menu   ( Menu * menu );

// map with all menu entries
map< string, Menu * > nodes;

// file containing menu description
static string menu_file   = "menu.desc";
static string header_file = "";
static string footer_file = "";
static string impressum   = "";

// nodes to print out
static TSLL< string > outnodes;

// which style to use
// 0 - default, with tables
// 1 - improved, less tables
// 2 - css
static int style = 2;

// css style file
static string css_file    = "";

// css style for layout
static string layout_css  = "css/layout.css";

// css style for tabs
static string tab_css     = "css/tabs.css";

// verbose mode
static bool verbose = false;

// create all nodes
static bool all_nodes = false;

//
// main function
//
int
main ( int argc, char ** argv )
{
    Menu  root( NULL, "root", "root", "", "" );

    nodes[ "root" ] = & root;

    parse_cmd( argc, argv );

#if 0
    read_menu( root, menu_file );
#else
    parse( & root, menu_file );
    adjust_menu( & root );
#endif
    
    if (verbose)
        root.print();

    // print all wished nodes
    if ( all_nodes )
    {
        TSLL< Menu * >             node_list;
        TSLL< Menu * >::TIterator  iter;

        node_list.append( & root );
        while ( node_list.size() > 0 )
        {
            root.clear_marks();
                
            if ( node_list.first_item()->name() != "root" )
                print_node( node_list.first_item() );
            
            iter = node_list.first_item()->submenus().first();
            while ( ! iter.eol() )
            {
                node_list.append( iter() );
                ++iter;
            }// while

            node_list.remove_first();
        }// while
    }// if
    else
    {
        TSLL< string >::TIterator  iter = outnodes.first();

        while ( ! iter.eol() )
        {
            root.clear_marks();
            print_node( iter() );
            ++iter;
        }// while
    }// else
}

//
// parse cmdline
//
static void
parse_cmd ( int argc, char ** argv )
{
#define OPTION(x)      strcmp( argv[i], (x) ) == 0
#define LOPTION(s,l)  (strcmp( argv[i], (s) ) == 0) || (strcmp( argv[i], (l) ) == 0)
    
    for ( int i = 1; i < argc; i++ )
    {
        if ( LOPTION( "-h", "--help" ) )                // help
            usage( 0 );
        else if ( OPTION( "--header" ) )                // header-file
        {
            if ( ++i < argc ) header_file = argv[i];
            else              usage( 1, "--header" );
        }// if
        else if ( LOPTION( "-f", "--footer" ) )         // footer-file
        {
            if ( ++i < argc ) footer_file = argv[i];
            else              usage( 1, "-f" );
        }// if
        else if ( LOPTION( "-m", "--menu" ) )           // menu-file
        {
            if ( ++i < argc ) menu_file = argv[i];
            else              usage( 1, "-m" );
        }// if
        else if ( LOPTION( "-v", "--verbose" ) )        // tree-file
            verbose = true;
        else if ( LOPTION( "-a", "--all" ) )            // all nodes
            all_nodes = true;
        else if ( LOPTION( "-s", "--style" ) )          // use css instead of pure tables
        {
            if ( ++i < argc ) style = atoi( argv[i] );
            else              usage( 1, "-s" );
        }// if
        else if ( LOPTION( "-c", "--cssfile" ) )        // css-layout file
        {
            if ( ++i < argc ) css_file = argv[i];
            else              usage( 1, "-c" );
        }
        else if ( LOPTION( "-i", "--impressum" ) )      // impressum-link
        {
            if ( ++i < argc ) impressum = argv[i];
            else              usage( 1, "-i" );
        }
        else
            outnodes.append( argv[i] );
    }// for
}

//
// print usage info
//
static void
usage ( int status, const string & param )
{
    if ( status != 0 )
    {
        if (param != "")
            cout << "wrong or missing parameter (" << param << "), try -h" << endl;
        else
            cout << "wrong or missing parameter, try -h" << endl;
            
        exit( status );
    }// if
    
    cout << "webgen v0.4.1 by Ronald Kriemann" << endl
         << "usage : webgen [ options ] node1 node2 ..." << endl
         << "where options include :" << endl
         << "    -a | --all              : create html-files for all nodes" << endl
         << "    -c | --cssfile <file>   : use <file> as css-layout (default=none)" << endl
         << "    -f | --footer <file>    : use footer given in <file> (default=none)" << endl
         << "    -i | --impressum <link> : use <link> as impressum (default=none)" << endl
         << "    -h | --help             : print this info" << endl
         << "         --header <file>    : use header given in <file> (default=none)" << endl
         << "    -m | --menu <file>      : use menu in <file>" << endl
         << "    -s | --style <int>      : which style to use (default=2)" << endl
         << "    -v | --verbose          : verbosity on" << endl;
    
    exit( status );
}

//
// adjust menu entries, copy data from father to sons/tabs
//
static void
adjust_sub ( Menu * father, Menu * sub )
{
    TSLL< std::string >::TIterator  iter;
    
    // inherit title, etc. if not set
    if ( sub->title()    == "" ) sub->title()    = father->title() + " : " + sub->name();
    if ( sub->headline() == "" ) sub->headline() = father->headline();
    if ( sub->header()   == "" ) sub->header()   = father->header();
    if ( sub->footer()   == "" ) sub->footer()   = father->footer();
    
    // prepend css-files of father to sons
    if ( ! sub->with_new_css() )
    {
        iter = father->css_files().first();
        
        while ( ! iter.eol() )
        {
            sub->css_files().prepend( iter() );
            ++iter;
        }// while
    }// if

    if (verbose)
    {
        cout << "node = (" << sub->name() << ")" << endl
             << "  father    = (" << sub->father()->name() << ")" << endl
             << "  content   = (" << sub->content() << ")" << endl
             << "  html_file = (" << sub->html_file() << ")" << endl
             << "  title     = (" << sub->title() << ")" << endl
             << "  headline  = (" << sub->headline() << ")" << endl
             << "  header    = (" << sub->header() << ")" << endl
             << "  footer    = (" << sub->footer() << ")" << endl;
        
        if ( sub->css_files().size() > 0 )
        {
            cout << "  cssfiles  = (";
            
            iter = sub->css_files().first();
            
            while ( ! iter.eol() > 0 )
            {
                cout <<  iter() << ", ";
                ++iter;
            }// while
            
            cout << ")" << endl;
        }// if
    }// if
}

static void
adjust_menu ( Menu * menu )
{
    TSLL< Menu * >::TIterator   iter = menu->submenus().first();
    
    while ( ! iter.eol() )
    {
        adjust_sub( menu, iter() );
        
        // recursive call 
        adjust_menu( iter() );
        
        ++iter;
    }// while

    iter = menu->tabs().first();

    while ( ! iter.eol() )
    {
        adjust_sub( menu, iter() );
        
        // recursive call 
        adjust_menu( iter() );
        
        ++iter;
    }// while
}

//
// write date into string
//
static std::string
getdate ()
{
    time_t      local_time;
    struct tm * timeval;
    char        buf[256];
    
    local_time = time( NULL );
    timeval = localtime( & local_time );

    snprintf( buf, sizeof(buf), "%04d-%02d-%02d",
              timeval->tm_year + 1900, timeval->tm_mday, timeval->tm_mon+1 );

    return std::string( buf );
}

//
// print single node
//
static void
print_node ( const string & id )
{
    if ( id == "" )
        return;
    
    //
    // go through menu and mark all fathers of node
    // and their direct submenus
    //

    Menu * node = nodes[ id ];

    if ( node == NULL )
    {
        cout << "print_node : could not find node with html-file = " << id << endl;
        return;
    }// if

    print_node( node );
}

static void
print_node ( Menu * node )
{
    if ( node == NULL )
        return;
    
    //
    // go through menu and mark all fathers of node
    // and their direct submenus
    //

    if ( verbose )
    {
        cout << "print_node : printing node " << node->name() << " ";
        
        if ( node->is_link() )
            cout << "(link)" << endl;
        else
            cout << "(" << node->html_file() << ")" << endl;
    }// if

    if ( node->is_link() )
        return;

    //
    // check if node has tabs and print them instead
    //
    
    if ( ! node->is_uniq() )
    {
        TSLL< Menu * >::TIterator  iter = node->tabs().first();

        while ( ! iter.eol() )
        {
            print_node( iter()->html_file() );
            ++iter;
        }// while

        return;
    }// if
    
    Menu                     * father;
    TSLL< Menu * >::TIterator  iter;

    node->set_mark( 1 );
    father = node;
    
    while ( true )
    {
        if ( father == NULL )
            break;
        
        father->set_mark( 1 );
        
        iter = father->submenus().first();
        while ( ! iter.eol() )
        {
//            iter()->set_mark( 1 );
            ++iter;
        }// while

        father = father->father();
    }// while

    //
    // now that we have marked all nodes in the menu, which
    // should be printed in the table, let's do it already
    //

    ofstream  file( node->html_file().c_str() );

    if ( ! file )
    {
        cout << "print_node : could not open output-file = " << node->html_file() << endl;
        return;
    }// if

    // first the header of the html-file
    if ( style == 0 )
    {
        file << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"" << endl
             << "\"http://www.w3.org/TR/REC-html40/loose.dtd\">" << endl
             << endl
             << "<html>" << endl
             << endl
             << "<head>" << endl
             << "  <title> " << node->title() << " </title>" << endl
             << "  <style type=text/css><!--" << endl
             << "    a:link    { color:#0000EE; }" << endl
             << "    a:visited { color:#6A6A6A; }" << endl
             << "    a:active  { color:#000000; }" << endl
             << "  --></style>" << endl
             << "</head>" << endl
             << endl
             << "<body bgcolor=\"#EED5B7\" text=\"black\"" << endl
             << "      style=\"font-family: Arial, Verdana, Helvetica, sans-serif;\">" << endl
             << "  <table width=\"100%\" border=0 cellspacing=0 cellpadding=0>" << endl;

        print_header( file, node );
        
        file << endl
             << "    <tr valign=top>" << endl
             << "      <td bgcolor=\"#F4C480\" width=\"20%\">" << endl;
        
        print_menu( file, nodes[ "root" ], node, "        " );
        
        file << "      </td>" << endl
             << "      <td width=\"80%\" bgcolor=\"white\">" << endl
             << "        <table width=\"100%\" cellspacing=0 cellpadding=8>" << endl
             << "          <tr>" << endl
             << "            <td width=\"100%\">" << endl
             << endl;
        
        // print the content
        print_content( file, node );
        
        file << endl
             << "            </td>" << endl
             << "          </tr>" << endl
             << "        </table>" << endl
             << "      </td>" << endl
             << "    </tr>" << endl
             << "    <tr> <td bgcolor=\"#CDB79E\" width=\"100%\" colspan=2> &nbsp; </td> </tr>"
             << endl;
        
        // and the footer
        print_footer( file, node );
        
        file << "    <tr>" << endl
             << "      <td width=\"100%\" colspan=2>" << endl
             << "        <table width=\"100%\" cellspacing=0 cellpadding=4>" << endl
             << "          <tr>" << endl
             << "            <td width=\"50%\" align=left>" << endl
             << "              <font size=1>" << endl
             << "                This page was last modified " << getdate() << " by " << endl
             << "                <a href=\"mailto:rok@mis.mpg.de\"> Ronald Kriemann </a>." << endl
             << "              </font>" << endl
             << "            </td>" << endl
             << "            <td width=\"50%\" align=right>" << endl
             << "              <font size=1>" << endl
             << "                <b> <a href=\"http://validator.w3.org/check/referer\">" << endl
             << "                Best viewed with any browser</a> </b>" << endl
             << "              </font>" << endl
             << "            </td>" << endl
             << "          </tr>" << endl
             << "        </table>" << endl
             << "      </td>" << endl
             << "    </tr>" << endl
             << "  </table>" << endl
             << endl
             << "</body>" << endl
             << "</html>" << endl;
    }// if
    else if ( style == 1 )
    {
        string  cssfile;

        file << "<?xml version=\"1.0\"?>" << endl
             << "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"" << endl
             << "    \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">" << endl
             << endl
             << "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">" << endl
             << endl
             << "<head>" << endl
             << "  <title> " << node->title() << " </title>" << endl;
        
        if ( css_file != "" )
            file << "  <link rel=\"stylesheet\" type=\"text/css\" href=\"" << css_file << "\" />"
                 << endl;
        else
        {
            TSLL< std::string >::TIterator  iter = node->css_files().first();

            while ( ! iter.eol() )
            {
                file << "  <link rel=\"stylesheet\" type=\"text/css\" href=\""
                     << rel_link( iter(), node->html_file() ) << "\" />" << endl;
                ++iter;
            }// while
        }// else
        
        file << "</head>" << endl
             << endl
             << "<body>" << endl
             << "  <table width=\"100%\" border=\"0\" cellspacing=\"0\" cellpadding=\"0\">"
             << endl;

        print_header( file, node );
        
        file << endl
             << "    <tr valign=\"top\">" << endl
             << "      <td  width=\"20%\" class=\"menu\">" << endl
             << "        &nbsp; <br />" << endl;
        
        print_menu( file, nodes[ "root" ], node, "        &nbsp; &nbsp;" );
        
        file << "        &nbsp; <br />" << endl
             << "      </td>" << endl
             << "      <td  width=\"80%\" class=\"content\">" << endl
             << "        <table width=\"100%\" cellspacing=\"0\" cellpadding=\"8\">" << endl
             << "          <tr>" << endl
             << "            <td width=\"100%\">" << endl
             << endl;
        
        // print the content
        print_content( file, node );
        
        file << endl
             << "            </td>" << endl
             << "          </tr>" << endl
             << "        </table>" << endl
             << "      </td>" << endl
             << "    </tr>" << endl;
        
        // and the footer
        print_footer( file, node );
        
        file << "    <tr>" << endl
             << "      <td width=\"100%\" colspan=\"2\">" << endl
             << "        <table width=\"100%\" cellspacing=\"0\" cellpadding=\"4\">" << endl
             << "          <tr>" << endl
             << "            <td width=\"50%\" align=\"left\">" << endl
             << "              <font size=\"1\">" << endl
             << "                This page was last modified " << getdate() << " by " << endl
             << "                <a href=\"mailto:rok@mis.mpg.de\"> Ronald Kriemann </a>." << endl
             << "              </font>" << endl
             << "            </td>" << endl
             << "            <td width=\"50%\" align=\"right\">" << endl
             << "              <font size=\"1\">" << endl
             << "                <b> <a href=\"http://validator.w3.org/check/referer\">" << endl
             << "                Best viewed with any browser</a> </b>" << endl
             << "              </font>" << endl
             << "            </td>" << endl
             << "          </tr>" << endl
             << "        </table>" << endl
             << "      </td>" << endl
             << "    </tr>" << endl
             << "  </table>" << endl
             << endl
             << "</body>" << endl
             << "</html>" << endl;
    }// if
    else if ( style == 2 )
    {
        string  cssfile;

        file << "<?xml version=\"1.0\"?>" << endl
             << "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\"" << endl
             << "    \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">" << endl
             << endl
             << "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">" << endl
             << endl
             << "<head>" << endl
             << "  <title> " << node->title() << " </title>" << endl;
        
        if ( css_file != "" )
            file << "  <link rel=\"stylesheet\" type=\"text/css\" href=\"" << css_file << "\" />"
                 << endl;
        else
        {
            TSLL< std::string >::TIterator  iter = node->css_files().first();

            file << "  <link rel=\"stylesheet\" type=\"text/css\" href=\""
                 << rel_link( layout_css, node->html_file() ) << "\" />"
                 << endl;
            
            while ( ! iter.eol() )
            {
                file << "  <link rel=\"stylesheet\" type=\"text/css\" href=\""
                     << rel_link( iter(), node->html_file() ) << "\" />" << endl;
                ++iter;
            }// while
        }// else

        // append tab-css style
        if ( node->is_tab() )
        {
            file << "  <link rel=\"stylesheet\" type=\"text/css\" href=\""
                 << rel_link( tab_css, node->html_file() ) << "\" />"
                 << endl;
        }// if
        
        file << "</head>" << endl
             << endl
             << "<body>" << endl
             << endl;

        print_header( file, node );
        
        file << endl
             << "<table class=\"columns\">" << endl
             << "<tr> <td class=\"left\">" << endl
             << endl;

        if ( node->is_tab() )
        {
            // add a line to align with tabbed content
            file << "<div class=\"menutab\"> &nbsp; </div>" << endl
                 << endl;
        }// if
        
        file << "<div class=\"menu\">" << endl;

        print_menu( file, nodes[ "root" ], node, "" );

        file << "</div>" << endl
             << endl
             << "</td>" << endl
             << endl
             << "<td class=\"right\">" << endl
             << endl;

        if ( node->is_tab() )
        {
            file << endl
                 << "<div class=\"tabs\">" << endl;
            
            print_tabs( file, nodes[ "root" ], node, "" );
            
            file << "</div>" << endl
                 << endl;
        }// if
        
        file << "<div class=\"content\">" << endl;
        
        // print the content
        print_content( file, node );

        file << "</div>" << endl
             << endl
             << "</td> </tr>" << endl
             << "</table>" << endl
             << endl;
        
        file << "<div class=\"copyright\">" << endl
             << "<table width=\"100%\" cellspacing=\"0\" cellpadding=\"4\">" << endl
             << "<tr>" << endl;

        if ( impressum != "" )
        {
            file << "  <td style=\"width: 50%; text-align: left;\">" << endl
                 << "    This page was last modified " << getdate() << " by" << endl
                 << "    <a href=\"mailto:rok@mis.mpg.de\"> Ronald Kriemann </a>." << endl
                 << "  </td>" << endl
                 << "  <td style=\"width: 20%; text-align: center;\">" << endl
                 << "    <a href=\"" << impressum << "\">Impressum</a>" << endl
                 << "  </td>" << endl
                 << "  <td style=\"width: 30%; text-align: right;\">" << endl
                 << "    <b> <a href=\"http://validator.w3.org/check/referer\">" << endl
                 << "    Best viewed with any browser</a> </b>" << endl
                 << "  </td>" << endl;
        }// if
        else
        {
            file << "  <td style=\"width: 50%; text-align: left;\">" << endl
                 << "    This page was last modified " << getdate() << " by" << endl
                 << "    <a href=\"mailto:rok@mis.mpg.de\"> Ronald Kriemann </a>." << endl
                 << "  </td>" << endl
                 << "  <td style=\"width: 50%; text-align: right;\">" << endl
                 << "    <b> <a href=\"http://validator.w3.org/check/referer\">" << endl
                 << "    Best viewed with any browser</a> </b>" << endl
                 << "  </td>" << endl;
        }// else

        file << "</tr>" << endl
             << "</table>" << endl
             << "</div>" << endl
             << endl
             << "</body>" << endl
             << "</html>" << endl;
    }// if
}

//
// print menu to html-file
//
static void
print_menu ( ostream & os, Menu * node, Menu * outnode, string ofs )
{
    if ( style == 0 )
    {
        if ( node->name() == "root" )
            os << ofs << "<table width=\"100%\" border=0 cellspacing=2 cellpadding=2 >" << endl
               << ofs << "  <tr>" << endl
               << ofs << "    <td> &nbsp; </td>" << endl
               << ofs << "    <td> &nbsp; </td>" << endl
               << ofs << "    <td> &nbsp; </td>" << endl
               << ofs << "  </tr>" << endl;
        else
            os << ofs << "<table>" << endl;
        
        TSLL< Menu * >::TIterator  iter = node->submenus().first();
        
        while ( ! iter.eol() )
        {
            os << ofs << "  <tr>" << endl
               << ofs << "    <td> &nbsp; </td>" << endl;
            
            if ((iter()->mark() == 0) || (iter()->nsubmenus() == 0))
            {
                os << ofs << "    <td> ";
                
                if ( iter() == outnode ) os << "<b>" << iter()->name() << "</b>";
                else                     os << "<a href=\"" << iter()->html_file() << "\">"
                                            << iter()->name() << "</a>";
                
                os << " </td>" << endl;
            }// if
            else
            {
                os << ofs << "    <td> " << endl;
                
                if ( iter() == outnode ) os << ofs << "      <b>" << iter()->name() << "</b> " << endl;
                else                     os << ofs << "      <a href=\"" << iter()->html_file()
                                            << "\">" << iter()->name() << "</a> " << endl;
                
                print_menu( os, iter(), outnode, ofs + "      " );
                
                os << ofs << "    </td>" << endl;
            }// else
            
            os << ofs << "    <td> &nbsp; </td>" << endl
               << ofs << "  </tr>" << endl;
            
            ++iter;
        }// while
        
        if ( node->name() == "root" )
            os << ofs << "  <tr>" << endl
               << ofs << "    <td> &nbsp; </td>" << endl
               << ofs << "    <td> &nbsp; </td>" << endl
               << ofs << "    <td> &nbsp; </td>" << endl
               << ofs << "  </tr>" << endl;
        
        os << ofs << "</table>" << endl;
    }// if
    else if ( style == 1 )
    {
        TSLL< Menu * >::TIterator  iter = node->submenus().first();

        while ( ! iter.eol() )
        {
            string test;

            if ( iter() == outnode ) os << ofs << "<b class=\"selected\">"
                                        << iter()->name() << "</b> <br />" << endl;
            else                     os << ofs << "<a href=\""
                                        << rel_link( iter()->html_file(),
                                                     outnode->html_file() ) << "\">"
                                        << iter()->name() << "</a> <br />" << endl;
            
            if ( ! ((iter()->mark() == 0) || (iter()->nsubmenus() == 0)) )
                print_menu( os, iter(), outnode, ofs + "&nbsp;&nbsp;" );
            
            ++iter;
        }// while
    }// if
    else if ( style == 2 )
    {
        TSLL< Menu * >::TIterator  iter = node->submenus().first();
        Menu                     * father = NULL;

        if ( outnode->is_tab() )
            father = outnode->father();
        
        os << ofs << "<ul>" << endl;
            
        while ( ! iter.eol() )
        {
            if (( iter() == outnode ) || (( father != NULL ) && ( iter() == father )))
                os << ofs << "  <li class=\"selected\"> " << iter()->name();
            else
            {
                os << ofs << "  <li class=\"item\">"
                   << "<a href=\"";

                if ( ! iter()->is_uniq() )
                    os << rel_link( iter()->tabs().first_item()->html_file(),
                                    outnode->html_file() )  << "\">";
                else
                    os << rel_link( iter()->html_file(),
                                    outnode->html_file() )  << "\">";
                    
                os << iter()->name() << "</a>";
            }// else
            
            if ( ! ((iter()->mark() == 0) || (iter()->nsubmenus() == 0)) )
            {
                os << endl;
                
                print_menu( os, iter(), outnode, ofs + "  " );
                
                os << ofs << "  </li>" << endl;
            }
            else
                os << "  </li>" << endl;
            
            ++iter;
        }// while
        
        os << ofs << "</ul>" << endl;
    }// if
}

//
// print tabs to html-file
//
static void
print_tabs ( ostream & os, Menu * node, Menu * outnode, string ofs )
{
    if ( style == 2 )
    {
        Menu * father = outnode->father();

        if ( father == NULL )
        {
            cout << "error: tab without father" << endl;
            return;
        }// if
        
        TSLL< Menu * >::TIterator  iter = father->tabs().first();
        
        while ( ! iter.eol() )
        {
            if ( iter() == outnode )
                os << ofs << "  <b> " << iter()->name() << " </b>" << endl;
            else
                os << ofs << "  <a class=\"plain\" href=\""
                   << rel_link( iter()->html_file(),
                                outnode->html_file() )  << "\">"
                   << iter()->name() << "</a>" << endl;
            
            ++iter;
        }// while
    }// if
}

//
// print the content of a node into stream
//
static void
print_content ( ostream & os, Menu * node )
{
    ifstream  is( node->content().c_str() );

    if ( ! is )
    {
        cout << "print_content : could not open file (" << node->content() << ")" << endl;
        return;
    }// if

    char ch;

    while ( is.get(ch) )
        os.put( ch );
}

//
// print header
//
static void
print_header ( ostream & os, Menu * node )
{
    string    file = (header_file == "" ? node->header() : header_file);
    ifstream  is( file.c_str() );

    if ( (file == "") || ! is )
    {
        if ( style == 0 )
        {
            // print standard header
            os << "    <tr>" << endl
               << "      <td bgcolor=\"#8B7D6B\" width=\"20%\"> &nbsp; </td>" << endl
               << "      <td bgcolor=\"#8B7D6B\" width=\"80%\"> &nbsp; </td>" << endl
               << "    </tr>" << endl
               << "    <tr> <td bgcolor=\"#8B7D6B\" width=\"100%\" align=center colspan=2>" << endl
               << "      <font size=\"+1\"> <b>" << node->headline() << "</b> </font>" << endl
               << "    </td> </tr>" << endl
               << "    <tr> <td bgcolor=\"#8B7D6B\" width=\"100%\" colspan=2> &nbsp; </td> </tr>" << endl
               << "    <tr> <td bgcolor=\"#CDB79E\" width=\"100%\" colspan=2> &nbsp; </td> </tr>" << endl;
        }// if
        else if ( style == 1 )
        {
            os << "    <tr>" << endl
               << "      <td class=\"style1\" width=\"20%\"> &nbsp; </td>" << endl
               << "      <td class=\"style2\" width=\"80%\" align=\"center\" height=\"60\">" << endl
               << "        <font size=\"+2\"> <b>"<< node->headline() << "</b> </font>" << endl
               << "      </td>" << endl
               << "    </tr>" << endl;
        }// if
        else if ( style == 2 )
        {
            if ( node->headline() != "" )
            {
                os << "<div class=\"header\">" << endl
                   << "  " << node->headline() << endl
                   << "</div>" << endl;
            }// if
        }// if
    
        return;
    }// if

    char ch;
    
    while ( is.get(ch) )
        os.put( ch );
}

//
// print footer
//
static void
print_footer ( ostream & os, Menu * node )
{
    string    file = (footer_file == "" ? node->footer() : footer_file);
    ifstream  is( file.c_str() );

    if ( (file == "") || ! is )
    {
        // print standard footer
        if ( style == 1 )
        {
            os << "    <tr>" << endl
               << "      <td class=\"style1\" width=\"20%\" > &nbsp; </td>" << endl
               << "      <td class=\"style2\" width=\"80%\" align=\"center\" height=\"30\">" << endl
               << "        &nbsp;" << endl
               << "      </td>" << endl
               << "    </tr>" << endl;
        }// if
        
        return;
    }// if

    char ch;

    while ( is.get(ch) )
        os.put( ch );
}

//
// return relative link from node2 to node1
//
static string
rel_link ( const string & file1, const string & file2 )
{
    string  link1 = file1;
    string  link2 = file2;
    string  link, d1, d2;

    // cut off common parent directories
    while ( true )
    {
        d1 = basedir( link1 );
        d2 = basedir( link2 );

        if ( d1 != d2 )
        {
            link1 = d1 + link1;
            link2 = d2 + link2;
            break;
        }// if

        if ( d1 == "" )
            break;
    }// while

    while ( basedir( link2 ) != "" )
        link += "../";

    link += link1;

    return link;
}

//
// return name of first parent directory
//
static string
basedir ( string & file )
{
    string    dir;
    unsigned  i;

    for ( i = 0; i < file.length(); i++ )
    {
        if ( file[i] == '/' )
        {
            for ( unsigned j = 0; j <= i; j++ )
                dir += file[j];

            file = & file.c_str()[i+1];

            return dir;
        }// if
    }// for

    return "";
}
